==> xmlroff-0.3.9.tar.gz: xmlroff-0.3.9/fo/fo-tree.c <==
/* Fo
 * fo-tree.c: Formatting object tree root
 *
 * Copyright (C) 2001 Sun Microsystems
 *
 * $Id: fo-tree.c,v 1.4 2005/11/20 13:43:07 tonygraham Exp $
 *
 * See Copying for the status of this software.
 */

#include "libfo/fo-utils.h"
#include "fo/fo-fo.h"
#include "fo/fo-fo-private.h"
#include "fo-tree.h"
#include "fo-node.h"
#include "fo-root.h"

enum {
  PROP_0
};

struct _FoTree
{
  FoFo parent_instance;

  GHashTable *master_name_hash;
  GHashTable *page_sequence_master_name_hash;
  FoFo *default_master;
};

struct _FoTreeClass
{
  FoFoClass parent_class;
  
};

static void fo_tree_init        (FoTree      *tree);
static void fo_tree_class_init  (FoTreeClass *klass);
static void fo_tree_finalize    (GObject           *object);
static void fo_tree_free_hash_key (gpointer key,
				   gpointer value,
				   gpointer user_data);
static void fo_tree_debug_dump_properties (FoFo *fo, gint depth);
static gboolean fo_tree_validate_content (FoFo *fo,
					  GError **error);

static gpointer parent_class;

/**
 * fo_tree_get_type:
 * 
 * Register the #FoTree object type.
 * 
 * Return value: #GType value of the #FoTree object type.
 **/
GType
fo_tree_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (FoTreeClass),
        (GBaseInitFunc) NULL,
        (GBaseFinalizeFunc) NULL,
        (GClassInitFunc) fo_tree_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (FoTree),
        0,              /* n_preallocs */
        (GInstanceInitFunc) fo_tree_init,
	NULL		/* value_table */
      };
      
      object_type = g_type_register_static (FO_TYPE_FO,
                                            "FoTree",
                                            &object_info, 0);
    }
  
  return object_type;
}

/**
 * fo_tree_init:
 * @tree: #FoTree object to initialise.
 * 
 * Implements #GInstanceInitFunc for #FoTree.
 **/
void
fo_tree_init (FoTree *tree)
{
  tree->master_name_hash = g_hash_table_new (g_str_hash, g_str_equal);
  tree->page_sequence_master_name_hash =
    g_hash_table_new (g_str_hash, g_str_equal);
  tree->default_master = NULL;
}

/**
 * fo_tree_class_init:
 * @klass: #FoTreeClass object to initialise.
 * 
 * Implements #GClassInitFunc for #FoTreeClass.
 **/
void
fo_tree_class_init (FoTreeClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  FoFoClass *fo_fo_class = FO_FO_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);
  
  object_class->finalize = fo_tree_finalize;

  fo_fo_class->debug_dump_properties = fo_tree_debug_dump_properties;
  fo_fo_class->validate_content = fo_tree_validate_content;
  /* No properties so no need to resolve property attributes. */
  fo_fo_class->resolve_property_attributes = NULL;
}

static void
fo_tree_finalize (GObject *object)
{
  FoTree *tree;

  tree = FO_TREE (object);

  g_hash_table_foreach (tree->master_name_hash,
			fo_tree_free_hash_key,
			NULL);
  g_hash_table_destroy (tree->master_name_hash);
  g_hash_table_foreach (tree->page_sequence_master_name_hash,
			fo_tree_free_hash_key,
			NULL);
  g_hash_table_destroy (tree->page_sequence_master_name_hash);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}


/**
 * fo_tree_new:
 * 
 * Creates a new #FoTree initialized to default value.
 * 
 * Return value: the new #FoTree
 **/
FoFo*
fo_tree_new (void)
{
  return FO_FO (g_object_new (fo_tree_get_type (), NULL));
}

static void
fo_tree_free_hash_key (gpointer key, gpointer value, gpointer user_data)
{
  g_free (key);
}

/**
 * fo_tree_master_name_add:
 * @tree: #FoTree to which to add a #FoPageMaster.
 * @name: Name to use for the #FoPageMaster.
 * @fo:   #FoPageMaster to add.
 * 
 * Adds @fo to @tree under @name.
 **/
void
fo_tree_master_name_add (FoFo *tree,
			 const gchar *name,
			 FoFo *fo)
{
  g_return_if_fail (tree != NULL);
  g_return_if_fail (FO_IS_TREE (tree));
  g_return_if_fail (name != NULL);
  g_return_if_fail (fo != NULL);
  g_return_if_fail (FO_IS_FO (fo));

  /*
  g_print ("Adding '%s'\n", name);
  */
  if (g_hash_table_lookup (FO_TREE (tree)->master_name_hash, name))
    {
      g_warning ("'%s' already in master name hash.", name);
    }
  else
    {
      g_hash_table_insert (FO_TREE (tree)->master_name_hash,
			   g_strdup (name),
			   fo);
    }

  if (FO_TREE (tree)->default_master == NULL)
    FO_TREE (tree)->default_master = fo;
}

/**
 * fo_tree_master_name_get:
 * @tree: #FoTree for which to get a #FoPageMaster.
 * @name: Name of the #FoPageMaster to get.
 * 
 * Gets the #FoPageMaster of @tree corresponding to @name.
 * 
 * Return value: #FoPageMaster matching @name, or NULL.
 **/
FoFo*
fo_tree_master_name_get (FoFo        *tree,
			 const gchar *name)
{
  g_return_val_if_fail (tree != NULL, NULL);
  g_return_val_if_fail (FO_IS_TREE (tree), NULL);
  g_return_val_if_fail (name != NULL, NULL);

  return g_hash_table_lookup (FO_TREE (tree)->master_name_hash,
			      name);
}

/**
 * fo_tree_page_sequence_master_name_add:
 * @tree: #FoTree for which to add a #FoPageSequenceMaster.
 * @name: Name of the #FoPageSequenceMaster.
 * @fo:   #FoPageSequenceMaster to add.
 * 
 * Adds @fo to to @tree.
 **/
void
fo_tree_page_sequence_master_name_add (FoFo *tree,
				       const gchar *name,
				       FoFo *fo)
{
  g_return_if_fail (tree != NULL);
  g_return_if_fail (FO_IS_TREE (tree));
  g_return_if_fail (name != NULL);
  g_return_if_fail (fo != NULL);
  g_return_if_fail (FO_IS_FO (fo));

  if (g_hash_table_lookup (FO_TREE (tree)->page_sequence_master_name_hash,
			   name))
    {
      g_warning ("'%s' already in page-sequence-master name hash.", name);
    }
  else
    {
      g_hash_table_insert (FO_TREE (tree)->page_sequence_master_name_hash,
			   g_strdup (name),
			   fo);
    }
}

/**
 * fo_tree_page_sequence_master_name_get:
 * @tree: #FoTree for which to get a named #FoPageSequenceMaster.
 * @name: Name of the #FoPageSequenceMaster.
 * 
 * Gets the #FoPageSequenceMaster of @tree with name matching @name.
 * 
 * Return value: #FoPageSequenceMaster corresponding to @name, or NULL.
 **/
FoFo*
fo_tree_page_sequence_master_name_get (FoFo *tree,
				       const gchar *name)
{
  g_return_val_if_fail (tree != NULL, NULL);
  g_return_val_if_fail (FO_IS_TREE (tree), NULL);
  g_return_val_if_fail (name != NULL, NULL);

  return g_hash_table_lookup (FO_TREE (tree)->page_sequence_master_name_hash,
			      name);
}

/**
 * fo_tree_default_master_get:
 * @tree: #FoTree for which to get the default master.
 * 
 * Gets the default page master of @tree.
 * 
 * Return value: The default page sequence master of @fo.
 **/
FoFo*
fo_tree_default_master_get (FoFo *tree)
{
  g_return_val_if_fail (tree != NULL, NULL);
  g_return_val_if_fail (FO_IS_TREE (tree), NULL);

  return FO_TREE (tree)->default_master;
}

static void
fo_tree_debug_dump_hash (gpointer key, gpointer value, gpointer data)
{
  gchar *indent = g_strnfill (GPOINTER_TO_INT (data) * 2, ' ');

  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "%s%s : %p",
	 indent, (gchar *) key, value);

  g_free (indent);
}

void
fo_tree_debug_dump_properties (FoFo *fo, gint depth)
{
  FoTree *fo_tree;
  gchar *indent = g_strnfill (depth * 2, ' ');

  g_return_if_fail (fo != NULL);
  g_return_if_fail (FO_IS_TREE (fo));

  fo_tree = FO_TREE (fo);


  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "%sdefault-master:   %p",
	 indent,
	 fo_tree->default_master);
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "%smaster-name hash:",
	 indent);
  g_hash_table_foreach (fo_tree->master_name_hash,
			fo_tree_debug_dump_hash,
			GINT_TO_POINTER (depth + 1));
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "%spage-sequence-master-name hash:",
	 indent);
  g_hash_table_foreach (fo_tree->page_sequence_master_name_hash,
			fo_tree_debug_dump_hash,
			GINT_TO_POINTER (depth + 1));

  g_free (indent);

  FO_FO_CLASS (parent_class)->debug_dump_properties (fo, depth + 2);
}

gboolean fo_tree_validate_content (FoFo *fo,
				   GError **error)
{
  GError *tmp_error;

  g_return_val_if_fail (FO_IS_TREE (fo), TRUE);

  if ((fo_node_n_children (FO_NODE (fo)) == 1) &&
      FO_IS_ROOT (fo_node_first_child (FO_NODE (fo))))
    return FALSE;
  else
    {
      tmp_error = g_error_new (FO_FO_ERROR,
			       FO_FO_ERROR_INVALID_CONTENT,
			       _(fo_fo_error_messages[FO_FO_ERROR_INVALID_CONTENT]),
			       fo_object_sprintf (FO_OBJECT (fo)));

      return fo_object_log_or_propagate_error (FO_OBJECT (fo),
					       error,
					       tmp_error);
    }
}

==> xmlroff-0.5.0.tar.gz: xmlroff-0.5.0/fo/fo-tree.c <==
/* Fo
 * fo-tree.c: Formatting object tree root
 *
 * Copyright (C) 2001 Sun Microsystems
 *
 * $Id: fo-tree.c,v 1.5 2006/03/21 23:23:14 tonygraham Exp $
 *
 * See Copying for the status of this software.
 */

#include "libfo/fo-utils.h"
#include "fo/fo-fo.h"
#include "fo/fo-fo-private.h"
#include "fo-tree.h"
#include "fo-node.h"
#include "fo-root.h"

enum {
  PROP_0
};

struct _FoTree
{
  FoFo parent_instance;

  GHashTable *id_hash;
  GHashTable *master_name_hash;
  GHashTable *page_sequence_master_name_hash;
  FoFo *default_master;
};

struct _FoTreeClass
{
  FoFoClass parent_class;
  
};

static void fo_tree_init        (FoTree      *tree);
static void fo_tree_class_init  (FoTreeClass *klass);
static void fo_tree_finalize    (GObject           *object);
static void fo_tree_free_hash_key (gpointer key,
				   gpointer value,
				   gpointer user_data);
static void fo_tree_debug_dump_properties (FoFo *fo, gint depth);
static gboolean fo_tree_validate_content (FoFo *fo,
					  GError **error);

static gpointer parent_class;

/**
 * fo_tree_get_type:
 * 
 * Register the #FoTree object type.
 * 
 * Return value: #GType value of the #FoTree object type.
 **/
GType
fo_tree_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (FoTreeClass),
        (GBaseInitFunc) NULL,
        (GBaseFinalizeFunc) NULL,
        (GClassInitFunc) fo_tree_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (FoTree),
        0,              /* n_preallocs */
        (GInstanceInitFunc) fo_tree_init,
	NULL		/* value_table */
      };
      
      object_type = g_type_register_static (FO_TYPE_FO,
                                            "FoTree",
                                            &object_info, 0);
    }
  
  return object_type;
}

/**
 * fo_tree_init:
 * @tree: #FoTree object to initialise.
 * 
 * Implements #GInstanceInitFunc for #FoTree.
 **/
void
fo_tree_init (FoTree *tree)
{
  tree->id_hash = g_hash_table_new (g_str_hash,
				    g_str_equal);
  tree->master_name_hash = g_hash_table_new (g_str_hash,
					     g_str_equal);
  tree->page_sequence_master_name_hash =
    g_hash_table_new (g_str_hash,
		      g_str_equal);
  tree->default_master = NULL;
}

/**
 * fo_tree_class_init:
 * @klass: #FoTreeClass object to initialise.
 * 
 * Implements #GClassInitFunc for #FoTreeClass.
 **/
void
fo_tree_class_init (FoTreeClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  FoFoClass *fo_fo_class = FO_FO_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);
  
  object_class->finalize = fo_tree_finalize;

  fo_fo_class->debug_dump_properties = fo_tree_debug_dump_properties;
  fo_fo_class->validate_content = fo_tree_validate_content;
  /* No properties so no need to resolve property attributes. */
  fo_fo_class->resolve_property_attributes = NULL;
}

static void
fo_tree_finalize (GObject *object)
{
  FoTree *tree;

  tree = FO_TREE (object);

  g_hash_table_foreach (tree->id_hash,
			fo_tree_free_hash_key,
			NULL);
  g_hash_table_destroy (tree->id_hash);
  g_hash_table_foreach (tree->master_name_hash,
			fo_tree_free_hash_key,
			NULL);
  g_hash_table_destroy (tree->master_name_hash);
  g_hash_table_foreach (tree->page_sequence_master_name_hash,
			fo_tree_free_hash_key,
			NULL);
  g_hash_table_destroy (tree->page_sequence_master_name_hash);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}


/**
 * fo_tree_new:
 * 
 * Creates a new #FoTree initialized to default value.
 * 
 * Return value: the new #FoTree
 **/
FoFo*
fo_tree_new (void)
{
  return FO_FO (g_object_new (fo_tree_get_type (), NULL));
}

static void
fo_tree_free_hash_key (gpointer key, gpointer value, gpointer user_data)
{
  g_free (key);
}

/**
 * fo_tree_id_add:
 * @tree: #FoTree to which to add a #FoFo.
 * @id:   Id to use for the #FoFo.
 * @fo:   #FoFo to add.
 * 
 * Adds @fo to @tree under @name.
 **/
void
fo_tree_id_add (FoFo        *tree,
		const gchar *id,
		FoFo        *fo)
{
  g_return_if_fail (tree != NULL);
  g_return_if_fail (FO_IS_TREE (tree));
  g_return_if_fail (id != NULL);
  g_return_if_fail (fo != NULL);
  g_return_if_fail (FO_IS_FO (fo));

  /*
  g_print ("Adding '%s'\n", id);
  */
  if (g_hash_table_lookup (FO_TREE (tree)->id_hash, id))
    {
      g_warning ("'%s' already in id hash.", id);
    }
  else
    {
      g_hash_table_insert (FO_TREE (tree)->id_hash,
			   g_strdup (id),
			   fo);
    }
}

/**
 * fo_tree_id_get:
 * @tree: #FoTree for which to get a #FoFo.
 * @id:   Id of the #FoFo to get.
 * 
 * Gets the #FoFo of @tree corresponding to @id.
 * 
 * Return value: #FoFo matching @id, or NULL.
 **/
FoFo*
fo_tree_id_get (FoFo        *tree,
		const gchar *id)
{
  g_return_val_if_fail (tree != NULL, NULL);
  g_return_val_if_fail (FO_IS_TREE (tree), NULL);
  g_return_val_if_fail (id != NULL, NULL);

  return g_hash_table_lookup (FO_TREE (tree)->id_hash,
			      id);
}

/**
 * fo_tree_master_name_add:
 * @tree: #FoTree to which to add a #FoPageMaster.
 * @name: Name to use for the #FoPageMaster.
 * @fo:   #FoPageMaster to add.
 * 
 * Adds @fo to @tree under @name.
 **/
void
fo_tree_master_name_add (FoFo *tree,
			 const gchar *name,
			 FoFo *fo)
{
  g_return_if_fail (tree != NULL);
  g_return_if_fail (FO_IS_TREE (tree));
  g_return_if_fail (name != NULL);
  g_return_if_fail (fo != NULL);
  g_return_if_fail (FO_IS_FO (fo));

  /*
  g_print ("Adding '%s'\n", name);
  */
  if (g_hash_table_lookup (FO_TREE (tree)->master_name_hash, name))
    {
      g_warning ("'%s' already in master name hash.", name);
    }
  else
    {
      g_hash_table_insert (FO_TREE (tree)->master_name_hash,
			   g_strdup (name),
			   fo);
    }

  if (FO_TREE (tree)->default_master == NULL)
    FO_TREE (tree)->default_master = fo;
}

/**
 * fo_tree_master_name_get:
 * @tree: #FoTree for which to get a #FoPageMaster.
 * @name: Name of the #FoPageMaster to get.
 * 
 * Gets the #FoPageMaster of @tree corresponding to @name.
 * 
 * Return value: #FoPageMaster matching @name, or NULL.
 **/
FoFo*
fo_tree_master_name_get (FoFo        *tree,
			 const gchar *name)
{
  g_return_val_if_fail (tree != NULL, NULL);
  g_return_val_if_fail (FO_IS_TREE (tree), NULL);
  g_return_val_if_fail (name != NULL, NULL);

  return g_hash_table_lookup (FO_TREE (tree)->master_name_hash,
			      name);
}

/**
 * fo_tree_page_sequence_master_name_add:
 * @tree: #FoTree for which to add a #FoPageSequenceMaster.
 * @name: Name of the #FoPageSequenceMaster.
 * @fo:   #FoPageSequenceMaster to add.
 * 
 * Adds @fo to to @tree.
 **/
void
fo_tree_page_sequence_master_name_add (FoFo *tree,
				       const gchar *name,
				       FoFo *fo)
{
  g_return_if_fail (tree != NULL);
  g_return_if_fail (FO_IS_TREE (tree));
  g_return_if_fail (name != NULL);
  g_return_if_fail (fo != NULL);
  g_return_if_fail (FO_IS_FO (fo));

  if (g_hash_table_lookup (FO_TREE (tree)->page_sequence_master_name_hash,
			   name))
    {
      g_warning ("'%s' already in page-sequence-master name hash.", name);
    }
  else
    {
      g_hash_table_insert (FO_TREE (tree)->page_sequence_master_name_hash,
			   g_strdup (name),
			   fo);
    }
}

/**
 * fo_tree_page_sequence_master_name_get:
 * @tree: #FoTree for which to get a named #FoPageSequenceMaster.
 * @name: Name of the #FoPageSequenceMaster.
 * 
 * Gets the #FoPageSequenceMaster of @tree with name matching @name.
 * 
 * Return value: #FoPageSequenceMaster corresponding to @name, or NULL.
 **/
FoFo*
fo_tree_page_sequence_master_name_get (FoFo *tree,
				       const gchar *name)
{
  g_return_val_if_fail (tree != NULL, NULL);
  g_return_val_if_fail (FO_IS_TREE (tree), NULL);
  g_return_val_if_fail (name != NULL, NULL);

  return g_hash_table_lookup (FO_TREE (tree)->page_sequence_master_name_hash,
			      name);
}

/**
 * fo_tree_default_master_get:
 * @tree: #FoTree for which to get the default master.
 * 
 * Gets the default page master of @tree.
 * 
 * Return value: The default page sequence master of @fo.
 **/
FoFo*
fo_tree_default_master_get (FoFo *tree)
{
  g_return_val_if_fail (tree != NULL, NULL);
  g_return_val_if_fail (FO_IS_TREE (tree), NULL);

  return FO_TREE (tree)->default_master;
}

static void
fo_tree_debug_dump_hash (gpointer key, gpointer value, gpointer data)
{
  gchar *indent = g_strnfill (GPOINTER_TO_INT (data) * 2, ' ');

  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "%s%s : %p",
	 indent, (gchar *) key, value);

  g_free (indent);
}

void
fo_tree_debug_dump_properties (FoFo *fo, gint depth)
{
  FoTree *fo_tree;
  gchar *indent = g_strnfill (depth * 2, ' ');

  g_return_if_fail (fo != NULL);
  g_return_if_fail (FO_IS_TREE (fo));

  fo_tree = FO_TREE (fo);


  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "%sdefault-master:   %p",
	 indent,
	 fo_tree->default_master);
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "%smaster-name hash:",
	 indent);
  g_hash_table_foreach (fo_tree->master_name_hash,
			fo_tree_debug_dump_hash,
			GINT_TO_POINTER (depth + 1));
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "%spage-sequence-master-name hash:",
	 indent);
  g_hash_table_foreach (fo_tree->page_sequence_master_name_hash,
			fo_tree_debug_dump_hash,
			GINT_TO_POINTER (depth + 1));

  g_free (indent);

  FO_FO_CLASS (parent_class)->debug_dump_properties (fo, depth + 2);
}

gboolean fo_tree_validate_content (FoFo *fo,
				   GError **error)
{
  GError *tmp_error;

  g_return_val_if_fail (FO_IS_TREE (fo), TRUE);

  if ((fo_node_n_children (FO_NODE (fo)) == 1) &&
      FO_IS_ROOT (fo_node_first_child (FO_NODE (fo))))
    return FALSE;
  else
    {
      tmp_error = g_error_new (FO_FO_ERROR,
			       FO_FO_ERROR_INVALID_CONTENT,
			       _(fo_fo_error_messages[FO_FO_ERROR_INVALID_CONTENT]),
			       fo_object_sprintf (FO_OBJECT (fo)));

      return fo_object_log_or_propagate_error (FO_OBJECT (fo),
					       error,
					       tmp_error);
    }
}
